/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon, Csaba Simon
 * Contributor(s): 
 */

#ifndef _CAROBMYSQL_HPP
#define _CAROBMYSQL_HPP

#include <config.h>

#include <CarobCommon.hpp>
#include <Converter.hpp>

/* MySQL includes */
#include <mysql_wrapper.h>

/* Carob includes */
#include <ConnectionParameters.hpp>
#include <Connection.hpp>
#include <DriverResultSet.hpp>
#include <Statement.hpp>

#include <string>
#include <vector>

#include <log4cxx/helpers/criticalsection.h>

const unsigned long MYSEQUOIA_MAGIC = 0xC00CA10B;

#define LOGGER_CONFIG_FILE "logger.conf"
#define MYSEQUOIA_CONFIG_FILE "mysequoia.conf"
#define DEFAULT_LOGGER_CONFIG_FILE CONFIGFILE_DIR"/"LOGGER_CONFIG_FILE
#define DEFAULT_MYSEQUOIA_CONFIG_FILE CONFIGFILE_DIR"/"MYSEQUOIA_CONFIG_FILE

class CarobStmt;

class CarobMYSQL : public CarobCommon
{
friend class CarobStmt;

public:
  CarobMYSQL (MYSQL * mysql, my_bool free_me);
  virtual ~CarobMYSQL ();

    /**
     * Connect to sequoia through Carob. Initialize the corresponding fields in the
     * MYSQL structure (eg. host, user, passwd, db, port, ...)
     * If db is null a connection will be a lazy connect which means when we specify
     * the database name with select_db the real connect will happen. This is needed
     * because sequoia does not support a connect without a virtual database name. 
     * @param host name of the host to connect
     * @param user name of the user which is used in connection
     * @param passwd password of the user used in connect
     * @param db name of the virtual database
     * @param port number of the port, if 0 use the default 25322
     * @param unix_socket if this parameter is not null, return an error
     * @param clientflag connection specific flags
     * @return if the connection was successful return true otherwise false
     *                  and set the corresponding mysql error
     */
  bool connect (const char *host, const char *user,
		const char *passwd, const char *db, unsigned int port,
		const char *unix_socket, unsigned long clientflag);

    /**
     * Modify the virtual database. If a lazy connection is ongoing then
     * try to connect to the database
     * @param db name of the virtual database
     * @return if the connection was successful return true otherwise false
     *                  and set the corresponding mysql error
     */
  bool select_db (const char *db);

    /**
     * Change the user or database. A default rollback will happen even if the
     * user is the same. If the new user can't connect to the server, the current
     * user will remain. If a lazy connection is ongoing then try to connect to
     * the database
     * @param user name of the user which is used in connection
     * @param passwd password of the user used in connect
     * @param db name of the virtual database
     * @return if the connection was successful return true otherwise false
     *                  and set the corresponding mysql error
     */
  bool change_user (const char *user, const char *passwd, const char *db);

    /**
     * Executes a query on the server. If a lazy connection is ongoing
     * returns an error. If a query was n
     * @param query the sql query to be executed, can contain null characters
     * @param length the length of the query parameter
     * @param parse set to true if the query need to be parsed
     * @return true if the query was executed otherwise false and set the
     *                  corresponding mysql error
     */
  bool real_query (const char *query, ulong length, bool parse = true);

  /**
   * Returns the number of warnings
   */
  unsigned int get_warning_count(void);

    /**
     * Gets the result of the query in MYSQL_RES format, and set the correspondent
     * MYSQL fields. If fetch_all is specified it will fetch all rows from the
     * server and close the query on the server, otherwise will fetch 
     * only 1 row and will leave open the query cursor on the server.
     * In this case at a new fetch row network operation will happen.
     * Only one live query / connection can exists.
     * @param fetch_all true if all results are storred in the memory
     * @return NULL if the query was execute only type (eq INSERT ...) or an error
     *                  occured, in this case sets the corresponding mysql error
     *         if fetch_all was true will return a pointer to the newly allocated
     *                  results. The results must be fread after use with free_results
     *         if fetch_all was false will return liveResultPtr
     */
  MYSQL_RES *get_results (my_bool fetch_all);

  /**
   * If a live query is ongoing return the next row from the result set.
   * there are more results on the server
   */
  MYSQL_ROW fetch_row ();

	/**
	 * In the case of multiple statement execution, returns true if
	 * there are more results on the server
	 */
  bool more_results ();

	/**
	 * In the case of multiple statement execution, fetch the next result from the
	 * server
   * @return 0 for success
   *        -1 if there are no more results
   *         1 if an error occured
   */
  int next_result ();

    /**
     * Free the stored results
     * If res == liveResultPtr will close the query on the server
     */
  static void free_results (MYSQL_RES * res);

    /**
     * Return true if there are no more rows to fetch from the result set or
     * there is no result set
     */
  bool eof ();

    /** 
     * Set the autocommit mode. If autocommit is on then all SQL statements will
     * be executed and commited as individual transactions.
     * @param mode - true set autocommit on
     *               false set autocommit off
     * @return true on success otherwise false
     */
  bool set_autocommit (const my_bool mode);

    /**
     * Make all changes permanent since the previous commit/rollback
     * @return true on success otherwise false
     */
  bool commit ();

    /**
     * Rollback all changes since the previous commit
     * @return true on succes otherwise false
     */
  bool rollback ();

    /**
     * List a table fields
     * @param  wild a regular expression mathing the field names
     * @param table name of the table
     * @return a MYSQL_RES with the name of the fields
     */
  MYSQL_RES *list_table_fields (const char *table, const char *wild);

    /**
     * List the databases names
     * @param wild a regular expression matching the databases names
     * @return MYSQL_RES with the name of the databases
     */
  MYSQL_RES *list_databases (const char *wild);

    /**
     * List the table names in the current database
     * @param wild a regular expression matching the tables names
     * @return MYSQL_RES with the name of the tables 
     */
  MYSQL_RES *list_tables (const char *wild);

  /**
   * Adds a command to the init command queue
   */
  void push_init_command(const char *cmd);
  
  /**
   * Set the character set for the connection
   */
  int set_character_set(const char *name);
  
  /**
   * Creates a prepared statement
   */
  CarobStmt *createPreparedStatment();
  
	/**
	 * Returns a pointer to the MYSQL structure
   */
  MYSQL *getMYSQL ();

  /**
   * Convert from wstring to string.
   * @param w the wstring.
   * @return a string.
   */
  virtual std::string from_wstring(const std::wstring &w);

  /**
   * Convert from string to wstring.
   * @param s the string.
   * @return a wstring.
   */
  virtual std::wstring to_wstring(const std::string &s);

  /**
   * Escapes special characters in a string for use in an SQL statement, 
   * taking into account the current charset of the connection.
   * @param to pointer to the converted string
   * @param from pointer to the string to be converted
   * @param length lenght of the string to convert
   * @return length of the new string if success
   *         -1 if an error occured
   */
  unsigned long real_escape_string (char *to, const char *from, unsigned long length);
  
protected:
  /**
   * Return a pointer to the last_error number in the mysql structure
   */
  virtual unsigned int *get_errno_ptr() { return &mysqlPtr->net.last_errno; }

  /**
   * Return a pointer to the last_error buffer in the mysql structure
   */
  virtual char *get_errmsg_ptr() { return mysqlPtr->net.last_error; }

  /**
   * Return a pointer to the sqlstate buffer in the mysql structure
   */
  virtual char *get_sqlstate_ptr() { return mysqlPtr->net.sqlstate; };

private:
  //pointer to a MYSQL structure which is connected to this class
  MYSQL * mysqlPtr;
  //pointer to a Connection pointer which is not null, if the connection is established
  CarobNS::Connection * connectionPtr;
  //pointer holding a carob statement, used to execute the commands
  CarobNS::Statement * stmtPtr;
  //pointer to a DriverResultSet containing the result of the executed select like statement
  CarobNS::DriverResultSet * drsPtr;
  //holds the result set for 'live' type of results
  MYSQL_RES *liveResultPtr;
  //list holding the command that need to be sent when a connection is established
  std::vector<std::string> init_cmd_list;
  //list holding all the prepared statements that this connection created
  std::vector<CarobStmt *> carobStmts;
  //critical section for accessing the carobStmts list
  log4cxx::helpers::CriticalSection cs_carobStmts;
  // character conversions
  Converter conv;
  // persistent connection
  bool usePersistentConnection;
  
  /**
   * Private default constructor.
   */
  CarobMYSQL ()
  {
  }

  /**
   * Allocate and fill the MYSQL structure
   * @param host name of the host to connect
   * @param user name of the user which is used in connection
   * @param passwd password of the user used in connect
   * @param db name of the virtual database
   * @param port number of the port
   */
  void set_connect_info (const char *host, const char *user,
			 const char *passwd, const char *db,
			 unsigned int port);

  /**
   * Delete and free the actual connection and it's associated resources
   * @param free_mysql defines if the internal mysql structure is destroyed
   */
  void delete_connection ();

  /**
   * Allocates a row from the freshly fetched results
   * @param row the resulting row
   * @param fields pointer to the field structure to update the max column length or null
   * @param field_count number of fields in the result
   * @param length total length of the result
   * @param lengths pointer to an array which will hold the length of each field
   */
  void alloc_row_data (MYSQL_ROW & row, MYSQL_FIELD * fields,
		       const unsigned int field_count, unsigned long &length,
		       unsigned long *lengths);

  /**
   * Release allocated row from the memory
   * @param row a pointer containing the row with data
   * @param field_count number of fields in the row
   */
  static void delete_row_data (MYSQL_ROW & row, const unsigned int field_count);
  
  /**
   * Read the MySQL ini file.
   */
   void read_ini_file();

  /**
   * Removes a prepared statement from the list
   */
  void removePreparedStatment(CarobStmt *stmt);
  
  /**
   * Remove this class from the prepared statements
   */
  void resetPreparedStatments();
};

#endif /* _CAROBMYSQL_HPP */
