/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon, Csaba Simon
 * Contributor(s): 
 */

#ifndef _UTILS_HPP
#define _UTILS_HPP

#include <string.h>
#include <string>

//MySQL includes
#include <mysql_wrapper.h>

extern const char *error_sqlstate[];

const int SQLT_OK = 0;
const int SQLT_UNKNOWN = 1;

typedef char *PCHAR;

#define FREE_AND_NULL(p) {delete (p); (p)=0;}
#define FREE_AND_NULL_ARRAY(p) {delete[] (p); (p)=0;}

#define MAX_CHARSET_LEN 40

/**
 * Creates a duplicate of a null terminated string and return the pointer to it.
 * @param src pointer to the source string
 * @return a pointer to the duplicate string
 */
char *cstrdup(const char *src);

/**
 * Creates a duplicate of the string of given length and return the pointer to it.
 * @param src pointer to the source string. Can contain null character.
 * @param len the length of the string
 * @return a pointer to the duplicate string
 */
char *cstrdup(const char *src, int len);

/**
 * Creates a duplicate of a null terminated string only if the destination
 * does not contain the same string and return the pointer to it.
 * @param dest pointer to the destination string. Can contain null character.
 * @param src pointer to the source string. Can contain null character.
 * @return a pointer to the duplicate string
 */
char *cstrdupcond(char * &dest, const char *src);

/**
 * Creates a duplicate of the string of given length only if the destination
 * does not contain the same string and return the pointer to it.
 * @param dest pointer to the destination string. Can contain null character.
 * @param src pointer to the source string. Can contain null character.
 * @param len the length of the string
 * @return a pointer to the duplicate string
 */
char *cstrdupcond(char * &dest, const char *src, int len);

/**
 * Returns a client error message associated with the error code.
 * @param code MySQL client error code
 * @return a the detailed error message or null if the error code was not a client error
 */
const char *get_error_msg(const int code);

/**
 * Converts the error code (usually returned by sequoia) to a mysql error code.
 * @param code sequoia error code
 * @return mapped MySQL error code
 */
int convert_errcode(const int code);

/**
 * Converts MySQL field type to sql type
 * @param mysql_type code of the MySQL field
 * @return mapped sql field type
 */
const int mysql_to_sql_type(enum enum_field_types mysql_type);

/**
 * Converts sql field type to mysql type
 * @param sqltype code of the sql field
 * @return mapped MySQL field type
 */
enum enum_field_types sql_to_mysql_type(const int sqltype);

/**
 * Return true if p is containing our Carob object
 */
bool isCarobObject(MYSQL *p);
bool isCarobObject(MYSQL_STMT *p);

/**
 * Data conversion functions between different MySQL data types
 */
void getFromTiny(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data);
void getFromShort(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data);
void getFromLong(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data);
void getFromLonglong(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data);
void getFromFloat(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data);
void getFromDouble(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data);
void getFromString(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data);
void getFromTime(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data);
void getFromDate(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data);
void getFromDateTime(MYSQL_BIND *bind, MYSQL_FIELD *field, void *data);

/**
 * Convert a MYSQL_TIME structure to a character string
 */
bool MYSQL_TIME_to_str(MYSQL_TIME *t, char *s, int maxlen, int fieldlen);

/**
 * Convert a character string to a MYSQL_TIME structure
 */
bool str_to_MYSQL_TIME(const char *s, MYSQL_TIME *t);

/**
 * compare the given strings ignore the leading whitespace and case
 */
bool string_match(const char *&src, const char *dst);

/**
 * get the next string (token) ignoring leading whitespace
 */
const char *get_next_string(const char *&src);

#endif /* _UTILS_HPP */
