/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon, Csaba Simon
 * Contributor(s): 
 */

#include <CarobMySQL.hpp>
#include <CarobStmt.hpp>
#include <IniParser.hpp>
#include <Utils.hpp>
#include <HostTokenizer.hpp>

/* Carob include */
#include <StringCodecs.hpp>
#include <ResultSetMetaData.hpp>

/* MySQL include */
#include <errmsg.h>

#include <cstdlib>
#include <string>
#include <algorithm>

using namespace CarobNS;
using namespace std;

CarobMYSQL::CarobMYSQL (MYSQL *mysql, my_bool free_me) :
   connectionPtr(0), stmtPtr(0), drsPtr(0), liveResultPtr(0), usePersistentConnection(true)
{
  LOG4CXX_DEBUG(logger, "Entering constructor; mysql=" << mysql << " free_me=" << (bool) free_me);
  static char version[6];
    
  /* initialize the mysql structure */
  memset(mysql, 0, sizeof(MYSQL));
  
  mysql->field_alloc.block_size = MYSEQUOIA_MAGIC;
  mysql->free_me = free_me;
  //TODO maybe a real server version string ???
  snprintf(version, sizeof(version), "%ud", MYSQL_VERSION_ID);
  mysql->server_version = version;
  mysql->protocol_version = PROTOCOL_VERSION;
  // TODO fill in other required mysql fields

  // force reading the my.cnf file
  mysql->options.my_cnf_group = cstrdup("client");

  // set the conversion to default "latin1"
  conv.set_code("latin1");
    
  mysqlPtr = mysql;

  LOG4CXX_DEBUG(logger, "Leaving constructor.");
}

CarobMYSQL::~CarobMYSQL ()
{
  LOG4CXX_DEBUG(logger, "Entering destructor.");
  
  resetPreparedStatments();

  delete_connection();

  LOG4CXX_INFO(logger, "Closing: host=" << mysqlPtr->host << " user=" << mysqlPtr->user << " passwd=" << mysqlPtr->passwd << 
        " db=" << mysqlPtr->db << " port=" << mysqlPtr->port);

  FREE_AND_NULL_ARRAY(mysqlPtr->host);
  FREE_AND_NULL_ARRAY(mysqlPtr->user);
  FREE_AND_NULL_ARRAY(mysqlPtr->passwd);
  FREE_AND_NULL_ARRAY(mysqlPtr->db);
  FREE_AND_NULL_ARRAY(mysqlPtr->host_info);
  FREE_AND_NULL_ARRAY(mysqlPtr->info);

  FREE_AND_NULL_ARRAY(mysqlPtr->options.my_cnf_file);
  FREE_AND_NULL_ARRAY(mysqlPtr->options.my_cnf_group);
  FREE_AND_NULL_ARRAY(mysqlPtr->options.host);
  FREE_AND_NULL_ARRAY(mysqlPtr->options.user);
  FREE_AND_NULL_ARRAY(mysqlPtr->options.password);
  FREE_AND_NULL_ARRAY(mysqlPtr->options.db);
  FREE_AND_NULL_ARRAY(mysqlPtr->options.charset_name);
  FREE_AND_NULL_ARRAY(mysqlPtr->options.charset_dir);
  mysqlPtr->options.port = 0;
    
  mysqlPtr->field_alloc.used = 0;
  mysqlPtr->extra_info = 0;

  LOG4CXX_DEBUG(logger, "Leaving destructor.");
}

bool
CarobMYSQL::connect (const char *host, const char *user,
		     const char *passwd, const char *db, unsigned int port,
		     const char *unix_socket, unsigned long clientflag)
{
  LOG4CXX_DEBUG(logger, "Entering connect: host=" << host << " user=" << user << " passwd=" << passwd << " db=" << db << " port=" << port);
  
  if (mysqlPtr->status != MYSQL_STATUS_READY)
  {
    set_error(CR_COMMANDS_OUT_OF_SYNC, SQLT_UNKNOWN);
    LOG4CXX_DEBUG(logger, "Leaving connect.");
    return false;
  }

  if (mysqlPtr->options.my_cnf_file || mysqlPtr->options.my_cnf_group)
  {
    read_ini_file();
    FREE_AND_NULL_ARRAY(mysqlPtr->options.my_cnf_file);
    FREE_AND_NULL_ARRAY(mysqlPtr->options.my_cnf_group);
  }

  if (unix_socket)
  {
    set_error(CR_SOCKET_CREATE_ERROR, SQLT_UNKNOWN);
    LOG4CXX_DEBUG(logger, "Leaving connect.");
    return false;
  }

  /* Must supply at least a user name, or connection to sequoia will fail */
  if (!user || !*user)
  {
    user = mysqlPtr->options.user;
    if (!user || !*user)
    {
      set_error(CR_NULL_POINTER, SQLT_UNKNOWN);
      LOG4CXX_DEBUG(logger, "Leaving connect.");
      return false;
    }
  }

  if (!host || !*host)
  {
    host = mysqlPtr->options.host;
    if (!host || !*host)
      host = "localhost";
  }
  
  if (!passwd)
  {
    passwd = mysqlPtr->options.password;
    if (!passwd)
      passwd = "";
  }
   
  if (!port)
  {
    port = mysqlPtr->options.port;
    if (!port)
      port = 25322;
  }
  
  try
  {
    /* set the default character set if not set */
    if (!mysqlPtr->options.charset_name || !*mysqlPtr->options.charset_name)
    {
      mysqlPtr->options.charset_name = cstrdup("latin1");
    }
    /* send it to the server */
    conv.set_code(mysqlPtr->options.charset_name);
    
    /* Check for delayed connection. If db name not given, the connection can't happen, because sequoia doesn't support it.
     * The real connection will happen after the user call the command mysql_select_db or mysql_real_connect */
    if (!db || !*db)
    {
      db = mysqlPtr->options.db;
      if (!db || !*db)
      {
        delete_connection();
        set_connect_info(host, user, passwd, 0, port);
    
        LOG4CXX_DEBUG(logger, "Leaving connect (delayed connection.)");
        return true;
      }
    }
    
    try 
    {
      std::vector<ControllerInfo> ctrls;

      HostTokenizer ht(host);
      HostTokenizer::const_iterator hti;
      
      for (hti = ht.begin(); hti != ht.end(); hti++)
      {
        ctrls.push_back(ControllerInfo(to_wstring(hti->host), hti->port));
        LOG4CXX_DEBUG(logger, "Connect (adding host): host=" << hti->host << " port=" << hti->port);
      }
      
      // we can retrieve the warnings only if we have persistent connection
      ConnectionParameters connectionParameters(ctrls, to_wstring(db), to_wstring(user), to_wstring(passwd), ht.getPolicy(), usePersistentConnection, usePersistentConnection?true:false);
      
      Connection *newConnectionPtr = new Connection(connectionParameters);
      
      if (newConnectionPtr)
      {
        //TODO handle not enough memory
        delete_connection();
        set_connect_info(host, user, passwd, db, port);
  
        LOG4CXX_INFO(logger, "Connected: host=" << host << " user=" << user << " passwd=" << passwd << " db=" << db << " port=" << port << " persistent_connection=" << usePersistentConnection);
        
        reset_error();
  
        connectionPtr = newConnectionPtr;
        
        //try to execute the initial commands after each connection
        std::vector<std::string>::const_iterator iter;
        for (iter=init_cmd_list.begin(); iter != init_cmd_list.end(); iter++)
          real_query(iter->data(), iter->length());
      }
  
      LOG4CXX_DEBUG(logger, "Leaving connect.");
      return true;
    } 
    catch (BackendException &e)
    {
      set_error(e);
      LOG4CXX_DEBUG(logger, "Leaving connect.");
      return false;
    }
    catch (CarobException &e)
    {
      set_error(e.getErrorCode(), from_wstring(e.description()).c_str(), from_wstring(e.getSQLState()).c_str());
      //TODO error handling
      LOG4CXX_DEBUG(logger, "Leaving connect.");
      return false;
    }
    catch (HostTokenizerException &e)
    {
      set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
      //TODO error handling
      LOG4CXX_DEBUG(logger, "Leaving connect.");
      return false;
    }
  }
  catch (ConverterException &e)
  {
    set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
    LOG4CXX_DEBUG(logger, "Leaving connect.");
    return false;
  }
}

bool
CarobMYSQL::select_db (const char *db)
{
  LOG4CXX_DEBUG(logger, "Entering select_db: db=" << db);

  bool result;
  
  /* Connect only if a user, password and host was already specified, otherwise raise an error */
  if (mysqlPtr->user && mysqlPtr->passwd && mysqlPtr->host)
  {
    result = connect(mysqlPtr->host, mysqlPtr->user, mysqlPtr->passwd, db, mysqlPtr->port, 0, mysqlPtr->client_flag);
    
  /*
   *  Do we need to send it out?
   *  the vdb name doesn't need to be the same as
   *  the db name
    if (result)
    {
      char buff[120];
      
      snprintf(buff, sizeof(buff), "use %s", db);
      result = real_query(buff, strlen(buff), false);
    }
  */

  }
  else
  {
    set_error(CR_NULL_POINTER, SQLT_UNKNOWN);
    result = false;
  }

  LOG4CXX_DEBUG(logger, "Leaving select_db: result=" << result);
  return result;
}

bool
CarobMYSQL::change_user (const char *user, const char *passwd, const char *db)
{
  LOG4CXX_DEBUG(logger, "Entering change_user: user=" << user << " passwd=" << " db=" << db);

  bool result = connect(mysqlPtr->host, user, passwd, db ? db : mysqlPtr->db, mysqlPtr->port, 0, mysqlPtr->client_flag);

  LOG4CXX_DEBUG(logger, "Leaving change_user: result=" << result);
  return result;
}

bool
CarobMYSQL::real_query (const char *query, ulong length, bool parse)
{
  LOG4CXX_DEBUG(logger, "Entering real_query: query=" << query << " length=" << length);

  LOG4CXX_INFO(logger, "Executing query=" << query << " length=" << length);

  if (mysqlPtr->status != MYSQL_STATUS_READY || !connectionPtr)
  {
    set_error(CR_COMMANDS_OUT_OF_SYNC, SQLT_UNKNOWN);
    LOG4CXX_DEBUG(logger, "Leaving real_query: result=0");
    return false;
  }

  // catch the "set names ..." queries and set the client encodings
  // do not send the query to the server
  if (parse)
  {
    const char *p = query, *q = 0;
    bool result = false;
    if (string_match(p, "set") && string_match(p, "names") && (q = get_next_string(p)) && (p - q < MAX_CHARSET_LEN))
    {
      char charset[MAX_CHARSET_LEN+1];
      strncpy(charset,q,p-q); charset[p-q]=0;
      result = set_character_set(charset) == 0 ? true : false;

      mysqlPtr->affected_rows = 0;
      mysqlPtr->field_count = 0;
      mysqlPtr->warning_count = 0;

      LOG4CXX_DEBUG(logger, "Leaving real_query: result=" << result);
      return result;
    }
  }

  try {  
    try {
      drsPtr = NULL;
  
      if (!stmtPtr)
      {
        stmtPtr = connectionPtr->createStatement();
        stmtPtr->setQueryTimeout(mysqlPtr->options.read_timeout);
        stmtPtr->setFetchSize(1);
      }
  
      FREE_AND_NULL_ARRAY(mysqlPtr->info);
      
      mysqlPtr->warning_count = 0;
  
      if (stmtPtr->execute(to_wstring(std::string(query,length))))
      {
        mysqlPtr->affected_rows = 0;
        drsPtr = stmtPtr->getResultSet();
        mysqlPtr->field_count = drsPtr ? drsPtr->getNumberOfColumns() : 0;
        mysqlPtr->status = MYSQL_STATUS_GET_RESULT;
        mysqlPtr->warning_count = get_warning_count();

        LOG4CXX_INFO(logger, "Executed query with result set. Field_count=" << mysqlPtr->field_count);
      }
      else
      {
        mysqlPtr->affected_rows = stmtPtr->getUpdateCount();
        mysqlPtr->status = MYSQL_STATUS_READY;
        mysqlPtr->field_count = 0;
        mysqlPtr->warning_count = get_warning_count();
        
        char info[100];
        snprintf(info, sizeof(info), "Records: %llu  Duplicates: 0  Warnings: %u", mysqlPtr->affected_rows, mysqlPtr->warning_count);
        info[sizeof(info)-1] = 0;
        mysqlPtr->info = cstrdup(info);
        LOG4CXX_INFO(logger, "Executed query. Affected_rows=" << mysqlPtr->affected_rows);
      }
  
      reset_error();
  
      LOG4CXX_DEBUG(logger, "Leaving real_query: result=1");
      return true;
    }
    catch (BackendException &e)
    {
      set_error(e);
      LOG4CXX_DEBUG(logger, "Leaving real_query: result=0.");
      return false;
    }
    catch (CarobException &e)
    {
      set_error(e.getErrorCode(), from_wstring(e.description()).c_str(), from_wstring(e.getSQLState()).c_str());
      LOG4CXX_DEBUG(logger, "Leaving real_query: result=0");
      return false;
    }
  }
  catch (ConverterException &e)
  {
    set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
    LOG4CXX_DEBUG(logger, "Leaving real_query: result=0");
    return false;
  }
}

unsigned int
CarobMYSQL::get_warning_count(void)
{
  LOG4CXX_DEBUG(logger, "Entering get_warning_count");

  unsigned int result = 0;

  // we can retrieve the warnings only if we have persistent connection
  if (stmtPtr && usePersistentConnection)
  {
    SQLWarning *warnings = stmtPtr->getWarnings();
     while (warnings)
     {
        result++;
        warnings = (SQLWarning *)warnings->getNext();
     }
  }
    
  LOG4CXX_DEBUG(logger, "Leaving get_warning_count: result=" << result);
  return result;
}

MYSQL_RES *
CarobMYSQL::get_results (my_bool fetch_all)
{
  LOG4CXX_DEBUG(logger, "Entering get_results: fetch_all=" << (bool) fetch_all);

  if (mysqlPtr->status != MYSQL_STATUS_GET_RESULT)
  {
  	//If the status is not get_result we don't throw cr_commands_out_of_sync
  	//this is required by PHP5
/*    set_error(CR_COMMANDS_OUT_OF_SYNC, SQLT_UNKNOWN);
    LOG4CXX_DEBUG(logger, "Leaving get_results: result=0");*/
    return 0;
  }

  try {
    try {
      if (stmtPtr)
      {
        MYSQL_RES *result = 0;
        
  //      drsPtr = stmtPtr->getResultSet();
        if (drsPtr)
        {
          //TODO handle not enough memory (std::nothrow)
          result = new MYSQL_RES;
          memset(result, 0, sizeof(MYSQL_RES));
  
          //Fill the fields metadata        
          ResultSetMetaData rsmd(drsPtr);
          result->field_count = get_query_fields(result->fields, &rsmd);
          result->handle = mysqlPtr;
          result->lengths = new unsigned long[result->field_count];
          
          //Fetch the rows if fetch_all was specified
          if (fetch_all)
          {
            result->data = new MYSQL_DATA;
            memset(result->data, 0, sizeof(MYSQL_DATA));
            result->data->fields = result->field_count;
            MYSQL_ROWS *prev = 0;
            //Fetch row by row and allocate memory for each row
            while (drsPtr->next())
            {
              MYSQL_ROWS *row = new MYSQL_ROWS;
              memset(row,0,sizeof(MYSQL_ROWS));
              alloc_row_data(row->data, result->fields, result->field_count, row->length, result->lengths);
              if (prev)
                prev->next = row;
              else
                result->data->data = row;
              prev = row;
              result->data->rows++;
            }
            result->row_count=result->data->rows;
            result->data_cursor = result->data->data;
            mysqlPtr->status = MYSQL_STATUS_READY;
      
            //Release the Carob dataset      
            //stmtPtr->close();
            drsPtr->close();
            drsPtr = 0;
          }
          else
          {
            //TODO check when to free the main internal liveResultPtr
            liveResultPtr = result;
            mysqlPtr->status = MYSQL_STATUS_USE_RESULT;
          }
  
          mysqlPtr->field_count = result->field_count;
          mysqlPtr->affected_rows = result->row_count;
          mysqlPtr->fields = result->fields;
  
          reset_error();
        }
  #ifdef MYSQL5
        else
          set_error(CR_NO_RESULT_SET, SQLT_UNKNOWN);
  #endif
  
        LOG4CXX_DEBUG(logger, "Leaving get_results: result=" << result);
        return result;
      }
      LOG4CXX_DEBUG(logger, "Leaving get_results: result=0");
      return 0;
    }
    catch (BackendException &e)
    {
      set_error(e);
      LOG4CXX_DEBUG(logger, "Leaving get_results: result=0.");
      return 0;
    }
    catch (CarobException &e)
    {
      set_error(e.getErrorCode(), from_wstring(e.description()).c_str(), from_wstring(e.getSQLState()).c_str());
      LOG4CXX_DEBUG(logger, "Leaving get_results: result=0");
      return 0;
    }
  }
  catch (ConverterException &e)
  {
    set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
    LOG4CXX_DEBUG(logger, "Leaving get_results: result=0");
    return 0;
  }
}

MYSQL_ROW
CarobMYSQL::fetch_row()
{
  LOG4CXX_DEBUG(logger, "Entering fetch_row.");

  if (mysqlPtr->status != MYSQL_STATUS_USE_RESULT || !drsPtr)
  {
    set_error(CR_COMMANDS_OUT_OF_SYNC, SQLT_UNKNOWN);
    LOG4CXX_DEBUG(logger, "Leaving fetch_row: result=0");
    return 0;
  }
  
  delete_row_data(liveResultPtr->row, liveResultPtr->field_count);
  if (drsPtr->next())
  {
    unsigned long len;
    alloc_row_data(liveResultPtr->row, liveResultPtr->fields, liveResultPtr->field_count, len, liveResultPtr->lengths);
    LOG4CXX_DEBUG(logger, "Leaving fetch_row: result=" << liveResultPtr->row);

    reset_error();

    return (liveResultPtr->current_row = liveResultPtr->row);
  }
  else
  {
    liveResultPtr->eof = 1;
    liveResultPtr->current_row = 0;
    LOG4CXX_DEBUG(logger, "Leaving fetch_row: result=0");

    reset_error();

    return 0;
  }
}

bool
CarobMYSQL::more_results ()
{
  LOG4CXX_DEBUG(logger, "Entering more_results.");

  bool result = stmtPtr ? stmtPtr->moreResults() : 0;

  LOG4CXX_DEBUG(logger, "Leaving more_results: result=" << result);
  return result;
}

int
CarobMYSQL::next_result ()
{
  LOG4CXX_DEBUG(logger, "Entering next_results.");

  int result = -1;
  
  if (stmtPtr)
  {
    if (liveResultPtr)
      free_results(liveResultPtr);
      
    try
    {
      try
      {
        if (stmtPtr->getMoreResults() || stmtPtr->getUpdateCount() != -1)
        {
          result = 0;
          mysqlPtr->affected_rows = stmtPtr->getUpdateCount();
          mysqlPtr->status = MYSQL_STATUS_GET_RESULT;
          drsPtr = stmtPtr->getResultSet();
        }
  
        reset_error();
      }
      catch (BackendException &e)
      {
        set_error(e);
        result = 1;
      }      
      catch (CarobException &e)
      {
        set_error(e.getErrorCode(), from_wstring(e.description()).c_str(), from_wstring(e.getSQLState()).c_str());
        result = 1;
      }
    }
    catch (ConverterException &e)
    {
      set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
      result = 1;
    }
  }
  
  LOG4CXX_DEBUG(logger, "Leaving next_results: result=" << result);
  return result;
}

void
CarobMYSQL::free_results (MYSQL_RES * res)
{
  LOG4CXX_DEBUG(logger, "Entering free_results: res=" << res);

  if (res)
  {
    if (res->unbuffered_fetch_cancelled == 1)
    {
      // delete only the resultset
      delete res;
    }
    else
    {
      // delete the resultset and all the fields 
      MYSQL *mysqlPtr = 0;
      CarobMYSQL *cm = 0;
      
      //Check if it is a "live" resultset
      if (!res->data)
      {
        mysqlPtr = res->handle;
  
  			//if it is a CarobMYSQL object, free the results from there too
        if (isCarobObject(mysqlPtr))
        {
          cm = (CarobMYSQL *) (mysqlPtr->field_alloc.used);
          if (res == cm->liveResultPtr)
          {
            cm->liveResultPtr = 0;
            cm->drsPtr = 0;
            cm->stmtPtr->close();
          }
        }
      }
      
      if (res->data)
      {
        MYSQL_ROWS *row = res->data->data;
        while (row)
        {
          MYSQL_ROWS *tempRow = row;
          row = row->next;
          delete_row_data(tempRow->data, res->field_count);
          delete tempRow;
        }
  
        delete res->data;
      }
      else
      {
        delete_row_data(res->current_row,res->field_count);
      }
      
      if (mysqlPtr && mysqlPtr->fields == res->fields)
      {
        mysqlPtr->fields = 0;
        mysqlPtr->field_count = 0;
      }
      delete_query_fields(res->fields, res->field_count);
  
      delete [] res->lengths;    
      delete res;
  
      if (mysqlPtr)
        mysqlPtr->status = MYSQL_STATUS_READY;
    }
  }
  LOG4CXX_DEBUG(logger, "Leaving free_results.");
}

bool
CarobMYSQL::eof ()
{
  LOG4CXX_DEBUG(logger, "Entering eof.");
  LOG4CXX_DEBUG(logger, "Leaving eof.");
  return 0;
}

bool
CarobMYSQL::set_autocommit (const my_bool mode)
{
  LOG4CXX_DEBUG(logger, "Entering set_autocommit: mode=" << (bool) mode);

  if (mysqlPtr->status != MYSQL_STATUS_READY || !connectionPtr)
  {
    set_error(CR_COMMANDS_OUT_OF_SYNC, SQLT_UNKNOWN);
    LOG4CXX_DEBUG(logger, "Leaving set_autocommit: result=0");
    return false;
  }
  
  try
  {
    try
    {
      connectionPtr->setAutoCommit(mode);
      LOG4CXX_DEBUG(logger, "Leaving set_autocommit: result=1");
  
      reset_error();
  
      return true;
    }
    catch (BackendException &e)
    {
      set_error(e);
      LOG4CXX_DEBUG(logger, "Leaving set_autocommit: result=0");
      return false;
    }          
    catch (CarobException &e)
    {
      set_error(e.getErrorCode(), from_wstring(e.description()).c_str(), from_wstring(e.getSQLState()).c_str());
      LOG4CXX_DEBUG(logger, "Leaving set_autocommit: result=0");
      return false;
    }
  }
  catch (ConverterException &e)
  {
    set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
    LOG4CXX_DEBUG(logger, "Leaving set_autocommit: result=0");
    return false;
  }
}

bool
CarobMYSQL::commit ()
{
  LOG4CXX_DEBUG(logger, "Entering commit.");

  if (mysqlPtr->status != MYSQL_STATUS_READY || !connectionPtr)
  {
    set_error(CR_COMMANDS_OUT_OF_SYNC, SQLT_UNKNOWN);
    LOG4CXX_DEBUG(logger, "Leaving commit: result=0");
    return false;
  }
  
  try
  {
    try
    {
      connectionPtr->commit();
      LOG4CXX_DEBUG(logger, "Leaving commit: result=1");
  
      reset_error();
  
      return true;
    }
    catch (BackendException &e)
    {
      set_error(e);
      LOG4CXX_DEBUG(logger, "Leaving commit: result=0");
      return false;
    }              
    catch (CarobException &e)
    {
      set_error(e.getErrorCode(), from_wstring(e.description()).c_str(), from_wstring(e.getSQLState()).c_str());
      LOG4CXX_DEBUG(logger, "Leaving commit: result=0");
      return false;
    }
  }
  catch (ConverterException &e)
  {
    set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
    LOG4CXX_DEBUG(logger, "Leaving commit: result=0");
    return false;
  }
}

bool
CarobMYSQL::rollback ()
{
  LOG4CXX_DEBUG(logger, "Entering rollback.");
  
  if (mysqlPtr->status != MYSQL_STATUS_READY || !connectionPtr)
  {
    set_error(CR_COMMANDS_OUT_OF_SYNC, SQLT_UNKNOWN);
    LOG4CXX_DEBUG(logger, "Leaving rollback: result=0");
    return false;
  }
  
  try
  {
    try
    {
      connectionPtr->rollback();
      LOG4CXX_DEBUG(logger, "Leaving rollback: result=1");
  
      reset_error();
  
      return true;
    }
    catch (BackendException &e)
    {
      set_error(e);
      LOG4CXX_DEBUG(logger, "Leaving rollback: result=0");
      return false;
    }
    catch (CarobException &e)
    {
      set_error(e.getErrorCode(), from_wstring(e.description()).c_str(), from_wstring(e.getSQLState()).c_str());
      LOG4CXX_DEBUG(logger, "Leaving rollback: result=0");
      return false;
    }
  }
  catch (ConverterException &e)
  {
    set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
    LOG4CXX_DEBUG(logger, "Leaving rollback: result=0");
    return false;
  }
}

MYSQL_RES *
CarobMYSQL::list_table_fields (const char *table, const char *wild)
{
  LOG4CXX_DEBUG(logger, "Entering list_table_fields: table=" << table << " wild=" << wild);

  MYSQL_RES *result;
  
  std::string query=std::string("SHOW FIELDS FROM ")+table;

  if (wild)
    query = query + " LIKE '"+wild+"'";
  if (real_query(query.c_str(),query.length()),false)
    result = get_results(true);
  else
    result = 0;

  LOG4CXX_DEBUG(logger, "Leaving list_table_fields: result=" << result);
  return result;
}

MYSQL_RES *
CarobMYSQL::list_databases (const char *wild)
{
  LOG4CXX_DEBUG(logger, "Entering list_databases: wild=" << wild);

  MYSQL_RES *result;

  std::string query="SHOW DATABASES";
  if (wild)
    query = query + " LIKE '"+wild+"'";
  if (real_query(query.c_str(),query.length()),false)
    result = get_results(true);
  else
    result = 0;

  LOG4CXX_DEBUG(logger, "Leaving list_databases: result=" << result);
  return result;
}

MYSQL_RES *
CarobMYSQL::list_tables (const char *wild)
{
  LOG4CXX_DEBUG(logger, "Entering list_tables: wild=" << wild);

  MYSQL_RES *result;
  
  std::string query="SHOW TABLES";
  if (wild)
    query = query + "LIKE '"+wild+"'";
  if (real_query(query.c_str(),query.length()),false)
    result = get_results(true);
  else
    result = 0;

  LOG4CXX_DEBUG(logger, "Leaving list_tables: result=" << result);
  return result;
}

void
CarobMYSQL::push_init_command(const char *cmd)
{
  LOG4CXX_DEBUG(logger, "Entering push_init_command: cmd=" << cmd);
  if (cmd && *cmd)
    init_cmd_list.push_back(string(cmd));
  LOG4CXX_DEBUG(logger, "Leaving push_init_command.");
}

int
CarobMYSQL::set_character_set(const char *name)
{
  LOG4CXX_DEBUG(logger, "Entering set_character_set: name=" << name);
  
  bool result = false;

  try
  {
    result = conv.set_code(name);
    
    FREE_AND_NULL_ARRAY(mysqlPtr->options.charset_name);
    mysqlPtr->options.charset_name = cstrdup(name);
  }
  catch (ConverterException &ex)
  {
    result=false;
  }

  if (!result)
    set_error(CR_CANT_READ_CHARSET, SQLT_UNKNOWN);
  else
    reset_error();

  LOG4CXX_DEBUG(logger, "Leaving set_character_set: result=" << result);
  return result == true ? 0 : 1;  
}

MYSQL *
CarobMYSQL::getMYSQL ()
{
  LOG4CXX_DEBUG(logger, "Entering getMYSQL.");
  LOG4CXX_DEBUG(logger, "Leaving getMYSQL.");
  return mysqlPtr;
}

void
CarobMYSQL::set_connect_info(const char *host, const char *user,
            const char *passwd, const char *db, unsigned int port)
{
  LOG4CXX_DEBUG(logger, "Entering set_connect_info: host=" << host << " user=" << user << " passwd=" << passwd << " db=" << db << " port=" << port);
  
  cstrdupcond(mysqlPtr->host, host);
  cstrdupcond(mysqlPtr->user, user);
  cstrdupcond(mysqlPtr->passwd, passwd);
  cstrdupcond(mysqlPtr->db, db);
  mysqlPtr->port = port;

  char buf[100];
  snprintf(buf, sizeof(buf), "%s via TCP/IP", host); buf[sizeof(buf)-1]='\0';
  cstrdupcond(mysqlPtr->host_info, buf);

  LOG4CXX_DEBUG(logger, "Leaving set_connect_info.");
}

void
CarobMYSQL::delete_connection()
{
  LOG4CXX_DEBUG(logger, "Entering delete_connection.");

  if (liveResultPtr)
  {
    free_results(liveResultPtr);
    FREE_AND_NULL(liveResultPtr);
  }
  //stmtPtr is freed by connection. just set our pointer to NULL
	stmtPtr = NULL;
  FREE_AND_NULL(connectionPtr);

  LOG4CXX_DEBUG(logger, "Leaving delete_connection.");
}

void
CarobMYSQL::alloc_row_data(MYSQL_ROW &row, MYSQL_FIELD *fields, const unsigned int field_count, unsigned long &length, unsigned long *lengths)
{
//Commented out because the debug info get unusable
//  LOG4CXX_DEBUG(logger, "Entering alloc_row_data: fields=" << fields << " field_count=" << field_count << " length=" << length << " lengths=" << lengths);

  /* TODO implement better
   * The main idea is to allocate a big enough row to hold all data as strings
   * separated by \0, and allocate an array of field_count+1 pointers, which
   * will point to this allocated row. The length of the data can be calculated
   * by substracting the pointer addresses. The last pointer is needed to
   * calculate the length of the last field */
  delete_row_data(row, field_count);
  
  row = new PCHAR[field_count+1];
  memset(row, 0, sizeof(PCHAR)*(field_count+1));
  
  std::string result, tmp;
  MYSQL_FIELD *fPtr = fields;
  int64_t lData;
  unsigned long field_len;

  /* TODO maybe we can change it, after carob implemented
   * date/time conversion functions correctly */
  //Set up a conversion buffer for time, date, datetime, float, double conversion
  char buffer[400];
  MYSQL_BIND b;

  memset(&b, 0, sizeof(b));
  b.buffer = &buffer;
  b.buffer_length = sizeof(buffer);
#ifdef MYSQL5
  b.is_null = &b.is_null_value;
  b.error = &b.error_value;
  b.length = &b.length_value;
#else
  b.is_null = &b.internal_is_null;
  b.length = &b.internal_length;
#endif
  b.buffer_type = MYSQL_TYPE_STRING;

  for (unsigned i=0; i<field_count; i++, fPtr++)
  {
    try
    {
      row[i] = (char *)(unsigned long)(result.size()+1);

      switch (fPtr->type)
      {
        case MYSQL_TYPE_TIME:
          lData = drsPtr->getAsInt(i+1);
          getFromTime(&b,fPtr,&lData);
          result += buffer;
          field_len = *b.length;
          break;
        case MYSQL_TYPE_DATE:
          lData = drsPtr->getAsInt64(i+1);
          getFromDate(&b,fPtr,&lData);
          result += buffer;
          field_len = *b.length;
          break;
        case MYSQL_TYPE_TIMESTAMP:
        case MYSQL_TYPE_DATETIME:
          {
            SQLTimeStamp ts = drsPtr->getTimeStamp(i+1);
            getFromDateTime(&b,fPtr,&ts);
            result += buffer;
            field_len = *b.length;
            break;
          }
        case MYSQL_TYPE_FLOAT:
          {
            float f = drsPtr->getFloat(i+1);
            getFromFloat(&b,fPtr,&f);
            result += buffer;
            field_len = strlen(buffer);
            break;
          }
        case MYSQL_TYPE_DOUBLE:
          {
            double d = drsPtr->getDouble(i+1);
            getFromDouble(&b,fPtr,&d);
            result += buffer;
            field_len = strlen(buffer);
            break;
          }
        default:
          if (fPtr->flags & BINARY_FLAG)
          {
            LargeData &d = drsPtr->getByteArray(i+1);
            tmp = string((char *)d.data, d.length);
          }
          else
            tmp = from_wstring(drsPtr->getAsString(i+1));
            
          field_len = tmp.length();
          result += tmp;
      }

      result.push_back(0);

      lengths[i] = field_len;      
      if (fields && fields[i].max_length < field_len)
        fields[i].max_length = field_len;
    }
    catch (NullValueException &e)
    {
      row[i] = 0;
    }
  }

  length = result.size();
  if (result.size())
  {
    char * presult = cstrdup(result.data(), result.size());
    for (unsigned i=0; i<field_count; i++)
      if (row[i])
        row[i] = presult + ((unsigned long) row[i] - 1);
    row[field_count] = presult + result.size();
  }

//  LOG4CXX_DEBUG(logger, "Leaving alloc_row_data.");
}

void
CarobMYSQL::delete_row_data(MYSQL_ROW &row, const unsigned int field_count)
{
//Commented out because the debug info get unusable
//  LOG4CXX_DEBUG(logger, "Entering delete_row_data: field_count=" << field_count);

  if (row)
  {
    for (unsigned int i=0; i<field_count; i++)
      if (row[i])
      {
        delete [] row[i];
        break;
      }

    delete [] row;
    row = 0;
  }

//  LOG4CXX_DEBUG(logger, "Leaving delete_row_data.");
}

void
CarobMYSQL::read_ini_file()
{
  LOG4CXX_DEBUG(logger, "Entering read_ini_file.");

  char *group;
  char *home_dir = getenv("HOME");
  char *conf_dir = getenv("LIBMYSEQUOIA_CONF_DIR");
  IniParser ini;

/* 
 * The order of reading the config files:
 * 1. LIBMYSEQUOIA_CONF_DIR env variable or /etc/mysequoia if not set
 * 2. "~/"
 * 3. ""
 * 4. user supplied options.my_cnf_file
 */

/* 1 */
  if (conf_dir)
  {
    string s = string(conf_dir) + "/" + MYSEQUOIA_CONFIG_FILE;
    ini.parseFile(s.c_str());
  }
  else
  {
    ini.parseFile(DEFAULT_MYSEQUOIA_CONFIG_FILE);
  }

/* 2 */
  if (home_dir)
  {
    string s = string(home_dir) + "/." + MYSEQUOIA_CONFIG_FILE;
    ini.parseFile(s.c_str());
  }

/* 3 */
  ini.parseFile(MYSEQUOIA_CONFIG_FILE);

/* 4 */
  if (mysqlPtr->options.my_cnf_file)
    ini.parseFile(mysqlPtr->options.my_cnf_file);

  if (mysqlPtr->options.my_cnf_group && *mysqlPtr->options.my_cnf_group)
    group = mysqlPtr->options.my_cnf_group;
  else
    group = "client";
  
  /*
   *  read the host, user, password, database, port, default-character-set, character-sets-dir
   *  from the config file
   */
  cstrdupcond(mysqlPtr->options.host, ini.get(group, "host").c_str());
  cstrdupcond(mysqlPtr->options.user, ini.get(group, "user").c_str());
  cstrdupcond(mysqlPtr->options.password, ini.get(group, "password").c_str());
  cstrdupcond(mysqlPtr->options.db, ini.get(group, "database").c_str());
  mysqlPtr->options.port = atoi(ini.get(group, "port").c_str());
  if (!mysqlPtr->options.charset_name || !*mysqlPtr->options.charset_name)
    cstrdupcond(mysqlPtr->options.charset_name, ini.get(group, "default-character-set").c_str());
  if (!mysqlPtr->options.charset_dir || !*mysqlPtr->options.charset_dir)
    cstrdupcond(mysqlPtr->options.charset_dir, ini.get(group, "character-sets-dir").c_str());

  /* read the init_command from the config file */
  push_init_command(ini.get(group, "init_command").c_str());

  /* set persistent connection */
  if (strncmp(ini.get(group, "persistent_connection").c_str(), "false", 5) == 0)
    usePersistentConnection = false;

  LOG4CXX_DEBUG(logger, "Leaving read_ini_file.");
}

CarobStmt *
CarobMYSQL::createPreparedStatment()
{
  LOG4CXX_DEBUG(logger, "Entering createPreparedStatment");
  CarobStmt *result = new CarobStmt(this);

  if (result)
  {
    cs_carobStmts.lock();
    carobStmts.push_back(result);
    cs_carobStmts.unlock();
  }

  LOG4CXX_DEBUG(logger, "Leaving createPreparedStatment result=" << result);
  return result;
}
  
void
CarobMYSQL::removePreparedStatment(CarobStmt *stmt)
{
  LOG4CXX_DEBUG(logger, "Entering deletePreparedStatment");
  cs_carobStmts.lock();
  vector<CarobStmt *>::iterator stmtIter = find(carobStmts.begin(), carobStmts.end(), stmt);
  if (stmtIter != carobStmts.end())
    carobStmts.erase(stmtIter);
  cs_carobStmts.unlock();
  LOG4CXX_DEBUG(logger, "Leaving deletePreparedStatment");
}

void
CarobMYSQL::resetPreparedStatments()
{
  LOG4CXX_DEBUG(logger, "Entering resetPreparedStatment");
  cs_carobStmts.lock();
  for (vector<CarobStmt *>::iterator stmtIter = carobStmts.begin(); stmtIter != carobStmts.end(); stmtIter++)
    (*stmtIter)->resetParent();
  cs_carobStmts.unlock();
  LOG4CXX_DEBUG(logger, "Leaving resetPreparedStatment");
}

std::string
CarobMYSQL::from_wstring(const wstring &w)
{
  return conv.from_wstring(w);
}

std::wstring
CarobMYSQL::to_wstring(const string &s)
{
  return conv.to_wstring(s);
}

unsigned long
CarobMYSQL::real_escape_string (char *to, const char *from, unsigned long length)
{
  if (length == 0)
  {
    if (to)
      *to = '\0';
      
    return 0;
  }
  
  try
  {
    wstring w = to_wstring(string(from, length));
    std::wstring::const_iterator i = w.begin();
    wstring r;
    wchar_t esc;
      
    for (; i != w.end(); i++)
    {
      esc = 0;
      switch (*i)
      {
        case L'\0': esc = L'0'; break; 
        case L'\n': esc = L'n'; break;
        case L'\r': esc = L'r'; break;
        case L'\\':
        case L'\'': 
        case L'"':  esc = *i; break;
        case L'\032': esc = L'Z'; break;
      }
      if (esc)
      {
        r.push_back(L'\\');
        r.push_back(esc);
      }
      else
      {
        r.push_back(*i);
      }
    }
  
    string s = from_wstring(r);
    
    size_t len = s.length() < 2*length ? s.length() : 2*length-1;
    memcpy(to, s.data(), len);
    to[len] = 0;
    
    return (s.length() < 2*length) ? (unsigned long) (s.length()) : (unsigned long) ~0;
  }
  catch (ConverterException &e)
  {
    set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
    return mysql_escape_string(to, from, length);
  }
}
