/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon, Csaba Simon
 * Contributor(s): 
 */

#include <CarobMySQL.hpp>
#include <Utils.hpp>

#include <string>

//Carob includes
#include <CarobStmt.hpp>
#include <Common.hpp>
#include <StringCodecs.hpp>
#include <TypeConstants.hpp>

//MySQL includes
#include <mysql_wrapper.h>
#include <my_alloc.h>
#include <errmsg.h>

using namespace CarobNS;
using namespace std;

#define DATA_NORMAL 1
#define DATA_TRUNCATED 2

CarobStmt::CarobStmt(CarobMYSQL *c): c_stmt(0), liveResultSet(0)
{
  LOG4CXX_DEBUG(logger, "Entering constructor: mysql=" << c);

  m_stmt = new MYSQL_STMT();
  memset(m_stmt, 0, sizeof(MYSQL_STMT));

  //Initialize the structure
  m_stmt->mem_root.used = (USED_MEM *) this;
  m_stmt->mem_root.block_size = MYSEQUOIA_MAGIC;
  m_stmt->mysql = c->getMYSQL();
#ifdef MYSQL5
  m_stmt->prefetch_rows = 1;
#endif
  m_stmt->state = MYSQL_STMT_INIT_DONE;

  cmysql = c;
  
  LOG4CXX_DEBUG(logger, "Leaving constructor.");
}

CarobStmt::~CarobStmt()
{
  LOG4CXX_DEBUG(logger, "Entering destructor.");

  if (cmysql)
    cmysql->removePreparedStatment(this);
    
  clear();
  
  FREE_AND_NULL(m_stmt);

  LOG4CXX_DEBUG(logger, "Leaving destructor.");
}
  
int
CarobStmt::prepare(const char *query, unsigned long length)
{
  LOG4CXX_DEBUG(logger, "Entering prepare: query=" << query << " length=" << length);

  if (!cmysql)
  {
    set_error(CR_SERVER_LOST, SQLT_UNKNOWN);
    LOG4CXX_DEBUG(logger, "Leaving prepare: result=1");
    return 1;
  }
  
  if (!cmysql->connectionPtr)
  {
    set_error(CR_COMMANDS_OUT_OF_SYNC, SQLT_UNKNOWN);
    LOG4CXX_DEBUG(logger, "Leaving prepare: result=1");
    return 1;
  }

  int result = 0;
  
  clear();

  try
  {   
    try
    {
      c_stmt = cmysql->connectionPtr->createParameterStatement(to_wstring(string(query, length)));
      m_stmt->param_count = c_stmt->getParameterCount();
      m_stmt->state = MYSQL_STMT_PREPARE_DONE;
      m_stmt->params = new MYSQL_BIND[m_stmt->param_count];
      bzero(m_stmt->params, m_stmt->param_count * sizeof(MYSQL_BIND));
      
      try
      {
        ResultSetMetaData *rsmdPtr = c_stmt->getMetaData();
        if (rsmdPtr)
        {
          m_stmt->field_count = rsmdPtr->getColumnCount();
          get_query_fields(m_stmt->fields, rsmdPtr);
          m_stmt->bind = new MYSQL_BIND[m_stmt->field_count];
        }
      } catch (BackendException &e)
      {
        LOG4CXX_DEBUG(logger, "Prepare: errno=" << e.getErrorCode() << " error=" << from_wstring(e.description()));
      }
    }
    catch (BackendException &e)
    {
      set_error(e);
      result = 1;
    }
    catch (CarobException &e)
    {
      set_error(e.getErrorCode(), from_wstring(e.description()).c_str(), from_wstring(e.getSQLState()).c_str());
      result = 1;
    }
  }
  catch (ConverterException &e)
  {
    set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
    result = 1;
  }
  
  
  LOG4CXX_DEBUG(logger, "Leaving prepare: result=" << result);
  
  return result;
}

bool
CarobStmt::bind_param(const MYSQL_BIND *bind)
{
  LOG4CXX_DEBUG(logger, "Entering bind_param: bind=" << bind);
 
  bool result = false;

  if (!m_stmt->param_count)
  {
    if (m_stmt->state < MYSQL_STMT_PREPARE_DONE)
    {
      set_error(CR_NO_PREPARE_STMT, SQLT_UNKNOWN);
      result = true;
    }
  }
  else
  {
    memcpy(m_stmt->params, bind, sizeof(MYSQL_BIND) * m_stmt->param_count);
    MYSQL_BIND *p = m_stmt->params;
    for (unsigned int no=1; no <= m_stmt->param_count && !result; no++, p++)
    {
      if (!p->is_null)
      {
#ifdef MYSQL5
        p->is_null = &p->is_null_value;
#else
        p->is_null = &p->internal_is_null;
#endif
        *p->is_null = false;
      }

      p->long_data_used = 0;
#ifdef MYSQL5
      p->row_ptr = 0;
#else
      p->inter_buffer = 0;
#endif
      p->param_number = no-1;
      
      switch (p->buffer_type)
      {
        case MYSQL_TYPE_TINY:
        case MYSQL_TYPE_SHORT:
        case MYSQL_TYPE_LONG:
        case MYSQL_TYPE_LONGLONG:
        case MYSQL_TYPE_FLOAT:
        case MYSQL_TYPE_DOUBLE:
        case MYSQL_TYPE_TINY_BLOB:
        case MYSQL_TYPE_MEDIUM_BLOB:
        case MYSQL_TYPE_LONG_BLOB:
        case MYSQL_TYPE_BLOB:
        case MYSQL_TYPE_VAR_STRING:
        case MYSQL_TYPE_STRING:
        case MYSQL_TYPE_TIME:
        case MYSQL_TYPE_DATE:
        case MYSQL_TYPE_DATETIME:
        case MYSQL_TYPE_TIMESTAMP:
        case MYSQL_TYPE_DECIMAL:
#ifdef MYSQL5        
        case MYSQL_TYPE_VARCHAR:
        case MYSQL_TYPE_NEWDECIMAL:
#endif
        case MYSQL_TYPE_NULL:
          break;
        default:
          set_error(CR_UNSUPPORTED_PARAM_TYPE, SQLT_UNKNOWN);
          result = true;
      }
    }

    if (!result)
      m_stmt->bind_param_done= true;
  }
  
  LOG4CXX_DEBUG(logger, "Leaving bind_param: result=" << result);
  return result;
}

bool
CarobStmt::send_long_data (unsigned int parameter_number, const char *data, unsigned long length)
{
  LOG4CXX_DEBUG(logger, "Entering send_long_data: parameter_number=" << parameter_number << " data=" << data << " length=" << length);

  bool result = true;
  //Because Carob doesn't support this feature it must be emulated. All data is collected in a string type
  //contained by the pointer row_ptr.
  if (!m_stmt->bind_param_done)
    set_error(CR_COMMANDS_OUT_OF_SYNC, SQLT_UNKNOWN);
  else
    if (parameter_number >= m_stmt->param_count)
      set_error(CR_INVALID_PARAMETER_NO, SQLT_UNKNOWN);
    else
    {
      MYSQL_BIND *bind = m_stmt->params + parameter_number;
      if (bind->buffer_type >= MYSQL_TYPE_TINY_BLOB &&
          bind->buffer_type <= MYSQL_TYPE_STRING)
      {
        //if we are at the first call, then allocate the string
        if (!bind->long_data_used)
        {
          bind->long_data_used = 1;
#ifdef MYSQL5          
          bind->row_ptr = (unsigned char *)new string;
#else
          bind->inter_buffer = (unsigned char *)new string;
#endif
        }
#ifdef MYSQL5          
        ((string *)(bind->row_ptr))->append(data, length);
#else
        ((string *)(bind->inter_buffer))->append(data, length);
#endif
        result = false;
      }
      else
        set_error(CR_INVALID_BUFFER_USE, SQLT_UNKNOWN);
    }

  LOG4CXX_DEBUG(logger, "Leaving send_long_data: result=" << result);
  return result;
}
         
int
CarobStmt::execute()
{
  LOG4CXX_DEBUG(logger, "Entering execute.");
  
  int result = 1;
  
  if (!cmysql)
    set_error(CR_SERVER_LOST, SQLT_UNKNOWN);
  else if (m_stmt->state < MYSQL_STMT_PREPARE_DONE)
    set_error(CR_NO_PREPARE_STMT, SQLT_UNKNOWN);
  else
    if (m_stmt->param_count && !m_stmt->params)
      set_error(CR_PARAMS_NOT_BOUND, SQLT_UNKNOWN);
    else
    {
#ifdef MYSQL5
      set_fetch_size(m_stmt->prefetch_rows);
#else
      set_fetch_size(1);
#endif

      try
      {
        try
        {
          MYSQL_BIND *p = m_stmt->params;
          for (unsigned int no=1; no <= m_stmt->param_count; no++, p++)
          {
            if (*p->is_null)
              c_stmt->setNull(no, mysql_to_sql_type(p->buffer_type));
            else
              switch (p->buffer_type)
              {
                case MYSQL_TYPE_TINY:
                  c_stmt->setByte(no, *(java_byte *)p->buffer);
                  break;
                case MYSQL_TYPE_SHORT:
                  c_stmt->setShort(no, *(short *)p->buffer);
                  break;
                case MYSQL_TYPE_LONG:
                  c_stmt->setInt(no, *(int *)p->buffer);
                  break;
                case MYSQL_TYPE_LONGLONG:
                  c_stmt->setLong(no, *(int64_t *)p->buffer);
                  break;
                case MYSQL_TYPE_FLOAT:
                  c_stmt->setFloat(no, *(float *)p->buffer);
                  break;
                case MYSQL_TYPE_DOUBLE:
                  c_stmt->setDouble(no, *(double *)p->buffer);
                  break;
                case MYSQL_TYPE_TINY_BLOB:
                case MYSQL_TYPE_MEDIUM_BLOB:
                case MYSQL_TYPE_LONG_BLOB:
                case MYSQL_TYPE_BLOB:
                  if (p->long_data_used && LONGDATAFIELD(p))
                  {
                    string *s = (string *)LONGDATAFIELD(p);
                    c_stmt->setBlob(no, s->data(), s->size());
                  }
                  else
                    c_stmt->setBlob(no, p->buffer, p->buffer_length);
                  break;
                case MYSQL_TYPE_DECIMAL:
  #ifdef MYSQL5
                case MYSQL_TYPE_NEWDECIMAL:
                case MYSQL_TYPE_VARCHAR:
  #endif
                case MYSQL_TYPE_VAR_STRING:
                case MYSQL_TYPE_STRING:
                  if (p->long_data_used && LONGDATAFIELD(p))
                    c_stmt->setString(no, to_wstring(*(string *)LONGDATAFIELD(p)));
                  else
                    // FIXME: can we use strnlen on string if the user does not set the length???
                    if (p->length)
                      c_stmt->setString(no, to_wstring(string((char *)p->buffer, *p->length)));
                    else
                      c_stmt->setString(no, to_wstring(string((char *)p->buffer, strnlen((char *)p->buffer, p->buffer_length))));
                  break;
                case MYSQL_TYPE_TIME:
                case MYSQL_TYPE_DATE:
                case MYSQL_TYPE_DATETIME:
                case MYSQL_TYPE_TIMESTAMP:
                  char buffer[20];
                  MYSQL_TIME_to_str((MYSQL_TIME *)p->buffer, buffer, sizeof(buffer), sizeof(buffer));
                  c_stmt->setString(no, to_wstring(string(buffer,strlen(buffer))));
                  break;
                case MYSQL_TYPE_NULL:
                  break;
                default:
                  LOG4CXX_FATAL(logger, "This should never happen!!");
                  /* TODO support other types */
              }
          }
          if (!c_stmt->execute())
            m_stmt->affected_rows = c_stmt->getUpdateCount();
          else
            liveResultSet = c_stmt->getResultSet();
  
          m_stmt->state = MYSQL_STMT_EXECUTE_DONE;
          result = 0;
        }
        catch (BackendException &e)
        {
          set_error(e);
        }        
        catch (CarobException &e)
        {
          set_error(e.getErrorCode(), from_wstring(e.description()).c_str(), from_wstring(e.getSQLState()).c_str());
        }
      }
      catch (ConverterException &e)
      {
        set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
      }
    }
  
  LOG4CXX_DEBUG(logger, "Leaving execute: result=" << result);
  return result;
}

bool
CarobStmt::bind_result(const MYSQL_BIND *bind)
{
  LOG4CXX_DEBUG(logger, "Entering bind_result: bind=" << bind);
 
  bool result = false;

  if (!m_stmt->field_count)
  {
    set_error(m_stmt->state < MYSQL_STMT_PREPARE_DONE ? CR_NO_PREPARE_STMT : CR_NO_STMT_METADATA, SQLT_UNKNOWN);
    result = true;
  }
  else
  {
    memcpy(m_stmt->bind, bind, sizeof(MYSQL_BIND) * m_stmt->field_count);
    MYSQL_BIND *p = m_stmt->bind;
    MYSQL_FIELD *fPtr = m_stmt->fields;
    for (unsigned int no=1; no <= m_stmt->field_count && !result; no++, p++, fPtr++)
    {
      p->param_number = no-1;
      p->offset = 0;

      /* set up the pointers if are not set up */
      link_bind(p);
      
      switch (p->buffer_type)
      {
        case MYSQL_TYPE_TINY:
          fPtr->max_length = 4;
          break;
        case MYSQL_TYPE_SHORT:
          fPtr->max_length = 6;
          break;
        case MYSQL_TYPE_LONG:
          fPtr->max_length = 11;
          break;
        case MYSQL_TYPE_LONGLONG:
          fPtr->max_length = 21;
          break;
        case MYSQL_TYPE_FLOAT:
          fPtr->max_length = 331;
          break;
        case MYSQL_TYPE_DOUBLE:
          fPtr->max_length = 331;
          break;
        case MYSQL_TYPE_TINY_BLOB:
        case MYSQL_TYPE_MEDIUM_BLOB:
        case MYSQL_TYPE_LONG_BLOB:
        case MYSQL_TYPE_BLOB:
        case MYSQL_TYPE_VAR_STRING:
        case MYSQL_TYPE_STRING:
        case MYSQL_TYPE_DECIMAL:
#ifdef MYSQL5
        case MYSQL_TYPE_VARCHAR:
        case MYSQL_TYPE_NEWDECIMAL:
#endif     
        case MYSQL_TYPE_TIME:
        case MYSQL_TYPE_DATE:
        case MYSQL_TYPE_DATETIME:
        case MYSQL_TYPE_TIMESTAMP:
        case MYSQL_TYPE_NULL:
          fPtr->max_length = 0;
          break;
        default:
          LOG4CXX_ERROR(logger, "bind_result: unsupported type pno=" << no << " type=" << p->buffer_type);
          set_error(CR_UNSUPPORTED_PARAM_TYPE, SQLT_UNKNOWN);
          result = true;
      }
    }

    if (!result)
#ifdef MYSQL5
      if (m_stmt->mysql->options.report_data_truncation)
        m_stmt->bind_result_done = DATA_TRUNCATED;
      else
#endif
        m_stmt->bind_result_done = DATA_NORMAL;
  }
  
  LOG4CXX_DEBUG(logger, "Leaving bind_result: result=" << result);
  return result;
}

int
CarobStmt::fetch()
{
  LOG4CXX_DEBUG(logger, "Entering fetch.");

  int result = 0;
#ifdef MYSQL5
  bool truncation = false;
#endif
  
  if (!cmysql && !m_stmt->result.data)
  {
    set_error(CR_SERVER_LOST, SQLT_UNKNOWN);
    result = 1;
  }
  else if (!m_stmt->bind_result_done || (m_stmt->state < MYSQL_STMT_EXECUTE_DONE))
  {
    set_error(CR_COMMANDS_OUT_OF_SYNC, SQLT_UNKNOWN);
    result = 1;
  }
  else
  try
  {
    //check if there are more results to fetch
    if (liveResultSet ? liveResultSet->next() : m_stmt->data_cursor != 0)
    {
      MYSQL_FIELD *fPtr = m_stmt->fields;
      MYSQL_BIND *bPtr = m_stmt->bind;
      for (unsigned colNo=0; colNo < m_stmt->field_count; colNo++, fPtr++, bPtr++)
      {
        fetch_field(fPtr, bPtr, colNo);

#ifdef MYSQL5
        truncation |= *bPtr->error;
#endif
      }
      
      if (m_stmt->result.data)
        m_stmt->data_cursor = m_stmt->data_cursor->next;
      else
      {
        m_stmt->result.rows++;
        m_stmt->affected_rows = m_stmt->result.rows;
      }
      
#ifdef MYSQL5
      if (truncation && (m_stmt->bind_result_done == DATA_TRUNCATED))
        result = MYSQL_DATA_TRUNCATED;
#endif

      m_stmt->state = MYSQL_STMT_FETCH_DONE;
    }
    else
      result = MYSQL_NO_DATA;
  }
  catch (ConverterException &e)
  {
    set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
    result = 1;
  }
  
  LOG4CXX_DEBUG(logger, "Leaving fetch: result=" << result);
  return result;
}

int
CarobStmt::fetch_column (MYSQL_BIND * bind, unsigned int column, unsigned long offset)
{
  LOG4CXX_DEBUG(logger, "Entering fetch_column: bind=" << bind << " column=" << column << " offset=" << offset);

  int result = 1;
  if (m_stmt->state != MYSQL_STMT_FETCH_DONE)
    set_error(CR_NO_DATA, SQLT_UNKNOWN);
  else
    if (column >= m_stmt->field_count)
      set_error(CR_INVALID_PARAMETER_NO, SQLT_UNKNOWN);
    else
      if (!bind)
        set_error(CR_NULL_POINTER, SQLT_UNKNOWN);
      else
      try
      {
        link_bind(bind);
        bind->offset = offset;

        fetch_field(m_stmt->fields+column, bind, column);
        
        result = 0;
      }
      catch (ConverterException &e)
      {
        set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
      }
      
  LOG4CXX_DEBUG(logger, "Leaving fetch_column: result=" << result);
  return result;
}

bool
CarobStmt::reset()
{
  LOG4CXX_DEBUG(logger, "Entering reset.");

  if (m_stmt->state >= MYSQL_STMT_PREPARE_DONE)
  {
    free_catched_result(m_stmt);
    free_long_data(m_stmt);
    if (c_stmt)
    {
      c_stmt->closeStatement();
      liveResultSet = 0;
    }
    m_stmt->state = MYSQL_STMT_PREPARE_DONE;
  }

  LOG4CXX_DEBUG(logger, "Leaving reset: result=0");
  return false;
}
  
MYSQL_RES *
CarobStmt::result_metadata()
{
  LOG4CXX_DEBUG(logger, "Entering result_metadata.");

  MYSQL_RES *result = 0;
  if (m_stmt->field_count)
  {
    result = new MYSQL_RES;
    memset(result, 0, sizeof(MYSQL_RES));
/*  FIXME double free or corruption!
    result->unbuffered_fetch_cancelled = 1; // reminder to not to free this resultset fields
    result->fields = m_stmt->fields;
    result->field_count = m_stmt->field_count;
*/
    result->data = new MYSQL_DATA;
    memset(result->data, 0, sizeof(MYSQL_DATA));
    //duplicate fields to the dummy resultset
    int field_count = m_stmt->field_count;
    result->fields = new MYSQL_FIELD[field_count];
    memcpy(result->fields, m_stmt->fields,  field_count * sizeof(MYSQL_FIELD));
    for (MYSQL_FIELD *colPtr=result->fields; field_count; colPtr++, field_count--)
    {
      colPtr->name = cstrdup(colPtr->name);
      colPtr->org_name = cstrdup(colPtr->org_name);
      colPtr->table = cstrdup(colPtr->table);
      colPtr->db = cstrdup(colPtr->db);
      colPtr->catalog = cstrdup(colPtr->catalog);
    }
    result->data->fields = result->field_count = m_stmt->field_count;
    result->eof = true;
  }  

  LOG4CXX_DEBUG(logger, "Leaving result_metadata: result=" << result);
  return result;
}

int
CarobStmt::store_result()
{
  LOG4CXX_DEBUG(logger, "Entering store_result.");

  int result = 0;
  
  if (!cmysql)
  {
    set_error(CR_SERVER_LOST, SQLT_UNKNOWN);
    result = 1;
  }
  if (m_stmt->state < MYSQL_STMT_EXECUTE_DONE)
  {
    set_error(CR_COMMANDS_OUT_OF_SYNC, SQLT_UNKNOWN);
    result = 1;
  }
  else
    try
    {
      try
      {
        if (m_stmt->field_count && liveResultSet)
        {    
          if (m_stmt->result.data)
            free_catched_result(m_stmt);
      
          memset(&m_stmt->result, 0, sizeof(MYSQL_DATA));
          m_stmt->result.fields = m_stmt->field_count;
          
          MYSQL_ROWS *prev = 0;
          //Fetch row by row and allocate memory for each row
          while (liveResultSet->next())
          {
            MYSQL_ROWS *row = new MYSQL_ROWS;
            memset(row,0,sizeof(MYSQL_ROWS));
            row->data = new PCHAR[m_stmt->field_count];
    
            MYSQL_FIELD *fPtr = m_stmt->fields;
            MYSQL_ROW dataPtr = row->data;
    
            for (unsigned colNo=1; colNo <= m_stmt->field_count; colNo++, fPtr++, dataPtr++)
              *dataPtr = (char *)alloc_fetch_field(fPtr, colNo);
            if (prev)
              prev->next = row;
            else
              m_stmt->result.data = row;
            prev = row;
            m_stmt->result.rows++;
          }
          m_stmt->data_cursor = m_stmt->result.data;
          m_stmt->affected_rows = m_stmt->result.rows;
      
          //Release the Carob dataset      
          c_stmt->closeStatement();
          liveResultSet = 0;
        }
      }
      catch (BackendException &e)
      {
        set_error(e);
        result = 1;
      }        
      catch (CarobException &e)
      {
        set_error(e.getErrorCode(), from_wstring(e.description()).c_str(), from_wstring(e.getSQLState()).c_str());
        result = 1;
      }
    }
    catch (ConverterException &e)
    {
      set_error(0, e.description().c_str(), error_sqlstate[SQLT_UNKNOWN]);
      result = 1;
    }

  LOG4CXX_DEBUG(logger, "Leaving store_result: result=" << result);
  return result;
}

bool
CarobStmt::free_result(MYSQL_STMT *stmt)
{
  LOG4CXX_DEBUG(logger, "Entering free_result: stmt=" << stmt);
  
  free_catched_result(stmt);

  LOG4CXX_DEBUG(logger, "Leaving free_result: result=0");
  return false;
}

void
CarobStmt::link_bind(MYSQL_BIND *bind)
{
  LOG4CXX_DEBUG(logger, "Entering link_bind: bind=" << bind);

#ifdef MYSQL5
  if (!bind->error)
    bind->error = &bind->error_value;
#endif
          
  if (!bind->is_null)
#ifdef MYSQL5
    bind->is_null = &bind->is_null_value;
#else
    bind->is_null = &bind->internal_is_null;
#endif
    
  if (!bind->length)
#ifdef MYSQL5
    bind->length = &bind->length_value;
#else
    bind->length = &bind->internal_length;
#endif

  LOG4CXX_DEBUG(logger, "Leaving link_bind");
}

void
CarobStmt::fetch_field(MYSQL_FIELD *fPtr, MYSQL_BIND *bPtr, unsigned colNo)
{
  LOG4CXX_DEBUG(logger, "Entering fetch_field: fPtr=" << fPtr << " bPtr=" << bPtr << " colNo=" << colNo);

  void *buffer = 0;

  //if there are cached results ...
  if (m_stmt->result.data)
    buffer = m_stmt->data_cursor->data[colNo];
  else
    buffer = alloc_fetch_field(fPtr, colNo+1);

  if (!buffer)
  {
    *bPtr->is_null = true;
  }
  else
  {
    *bPtr->is_null = false;
    switch (fPtr->type)
    {
      case MYSQL_TYPE_TINY:
        getFromTiny(bPtr, fPtr, buffer);
        break;
      case MYSQL_TYPE_SHORT:
        getFromShort(bPtr, fPtr, buffer);
        break;
      case MYSQL_TYPE_LONG:
        getFromLong(bPtr, fPtr, buffer);
        break;
      case MYSQL_TYPE_LONGLONG:
        getFromLonglong(bPtr, fPtr, buffer);
        break;
      case MYSQL_TYPE_FLOAT:
        getFromFloat(bPtr, fPtr, buffer);
        break;
      case MYSQL_TYPE_DOUBLE:
        getFromDouble(bPtr, fPtr, buffer);
        break;
      case MYSQL_TYPE_TINY_BLOB:
      case MYSQL_TYPE_MEDIUM_BLOB:
      case MYSQL_TYPE_LONG_BLOB:
      case MYSQL_TYPE_BLOB:
      case MYSQL_TYPE_VAR_STRING:
      case MYSQL_TYPE_STRING:
      case MYSQL_TYPE_DECIMAL:
#ifdef MYSQL5
      case MYSQL_TYPE_VARCHAR:
      case MYSQL_TYPE_NEWDECIMAL:
#endif
        if ((m_stmt->update_max_length) && (((string *)buffer)->length() > fPtr->max_length))
          fPtr->max_length = ((string *)buffer)->length();
        getFromString(bPtr, fPtr, buffer);
        break;
      case MYSQL_TYPE_TIME:
        getFromTime(bPtr, fPtr, buffer);
        break;
      case MYSQL_TYPE_DATE:
        getFromDate(bPtr, fPtr, buffer);
        break;
      case MYSQL_TYPE_DATETIME:
      case MYSQL_TYPE_TIMESTAMP:
        getFromDateTime(bPtr, fPtr, buffer);
        break;
      case MYSQL_TYPE_NULL:
        *bPtr->is_null = true;
        break;
      default:
        LOG4CXX_FATAL(logger, "This should never happen!!");
    }
  }

  if (!m_stmt->result.data)
    delete_fetch_field(fPtr, buffer);

  LOG4CXX_DEBUG(logger, "Leaving fetch_field.");
}

void *
CarobStmt::alloc_fetch_field(MYSQL_FIELD *fPtr, unsigned colNo)
{
  LOG4CXX_DEBUG(logger, "Entering alloc_fetch_field: fPtr=" << fPtr << " colNo=" << colNo);
  
  void *result = 0;
  
  try
  {
    switch (fPtr->type)
    {
      case MYSQL_TYPE_TINY:
        fPtr->max_length = 4;
        result = new unsigned char(liveResultSet->getInt32(colNo));
        break;
      case MYSQL_TYPE_SHORT:
        fPtr->max_length = 6;
        result = new short(liveResultSet->getInt32(colNo));
        break;
      case MYSQL_TYPE_LONG:
        fPtr->max_length = 11;
        //we need to use getAsInt and not getInt32 because mysql java client will convert the
        //unsigned integer to long
        result = new int(liveResultSet->getAsInt(colNo));
        break;
      case MYSQL_TYPE_LONGLONG:
        fPtr->max_length = 21;
        //we need to use getAsInt64 and not getInt64 because mysql java client will convert the
        //unsigned long to bigint
        if (fPtr->flags & UNSIGNED_FLAG)
          result = new int64_t(liveResultSet->getAsUInt64(colNo));
        else
          result = new int64_t(liveResultSet->getAsInt64(colNo));
        break;
      case MYSQL_TYPE_FLOAT:
        fPtr->max_length = 331;
        result = new float(liveResultSet->getFloat(colNo));
        break;
      case MYSQL_TYPE_DOUBLE:
        fPtr->max_length = 331;
        result = new double(liveResultSet->getDouble(colNo));
        break;
      case MYSQL_TYPE_TINY_BLOB:
      case MYSQL_TYPE_MEDIUM_BLOB:
      case MYSQL_TYPE_LONG_BLOB:
      case MYSQL_TYPE_BLOB:
      case MYSQL_TYPE_VAR_STRING:
      case MYSQL_TYPE_STRING:
      case MYSQL_TYPE_DECIMAL:
#ifdef MYSQL5
      case MYSQL_TYPE_VARCHAR:
      case MYSQL_TYPE_NEWDECIMAL:
#endif
        if (fPtr->flags & BINARY_FLAG)
        {
          LargeData &d = liveResultSet->getByteArray(colNo);
          fPtr->max_length = d.length;
          result = new string((char *)d.data, d.length);
        }
        else
        {
          std::wstring ws =liveResultSet->getAsString(colNo);
          fPtr->max_length = ws.length();
          result = new string(from_wstring(ws));
        }
        break;
      case MYSQL_TYPE_TIME:
        fPtr->max_length = 15;
        result = new int(liveResultSet->getAsInt(colNo));
        break;
      case MYSQL_TYPE_DATE:
        fPtr->max_length = 10;
        result = new int64_t(liveResultSet->getAsInt64(colNo));
        break;
      case MYSQL_TYPE_DATETIME:
      case MYSQL_TYPE_TIMESTAMP:
        fPtr->max_length = 30;
        result = new SQLTimeStamp(liveResultSet->getTimeStamp(colNo));
        break;
      case MYSQL_TYPE_NULL:
        fPtr->max_length = 0;
        break;
      default:
        LOG4CXX_FATAL(logger, "This should never happen!!");
    }
  }
  catch (NullValueException)
  {
  }
  catch (BackendException &e)
  {
    set_error(e);
  }        
  catch (CarobException &e)
  {
    set_error(e.getErrorCode(), from_wstring(e.description()).c_str(), from_wstring(e.getSQLState()).c_str());
  }
        
  LOG4CXX_DEBUG(logger, "Leaving alloc_fetch_field: result=" << result);
  return result;
}

void
CarobStmt::delete_fetch_field(MYSQL_FIELD *fPtr, void *buffer)
{
  LOG4CXX_DEBUG(logger, "Entering delete_fetch_field fPtr="<< fPtr << " type=" << fPtr->type << " buffer=" << buffer); 

  switch (fPtr->type)
  {
    case MYSQL_TYPE_TINY:
      delete (unsigned char *)buffer;
      break;
    case MYSQL_TYPE_SHORT:
      delete (short *)buffer;
      break;
    case MYSQL_TYPE_LONG:
    case MYSQL_TYPE_TIME:
      delete (int *)buffer;
      break;
    case MYSQL_TYPE_DATE:
    case MYSQL_TYPE_LONGLONG:
      delete (int64_t *)buffer;
      break;
    case MYSQL_TYPE_FLOAT:
      delete (float *)buffer;
      break;
    case MYSQL_TYPE_DOUBLE:
      delete (double *)buffer;
      break;
    case MYSQL_TYPE_TINY_BLOB:
    case MYSQL_TYPE_MEDIUM_BLOB:
    case MYSQL_TYPE_LONG_BLOB:
    case MYSQL_TYPE_BLOB:
    case MYSQL_TYPE_VAR_STRING:
    case MYSQL_TYPE_STRING:
    case MYSQL_TYPE_DECIMAL:
#ifdef MYSQL5
    case MYSQL_TYPE_VARCHAR:
    case MYSQL_TYPE_NEWDECIMAL:
#endif
      delete (string *)buffer;
      break;
    case MYSQL_TYPE_DATETIME:
    case MYSQL_TYPE_TIMESTAMP:
      delete (SQLTimeStamp *)buffer;
      break;
    case MYSQL_TYPE_NULL:
      break;
    default:
      LOG4CXX_FATAL(logger, "This should never happen!!");
  }

  LOG4CXX_DEBUG(logger, "Leaving delete_fetch_field.");
}

void
CarobStmt::free_catched_result(MYSQL_STMT *stmt)
{
  LOG4CXX_DEBUG(logger, "Entering free_catched_result: stmt=" << stmt);

  if (stmt->result.data)
  {
    MYSQL_ROWS *tmp;
    while ((tmp = stmt->result.data))
    {
      stmt->result.data = stmt->result.data->next;
      MYSQL_FIELD *fPtr = stmt->fields;
      MYSQL_ROW dataPtr = tmp->data;
      for (unsigned colNo=0; colNo < stmt->field_count; colNo++, fPtr++, dataPtr++)
        delete_fetch_field(fPtr, *dataPtr);
      delete []tmp->data;
      delete tmp;
    }
    stmt->result.rows= 0;
    stmt->data_cursor= NULL;
  }

  LOG4CXX_DEBUG(logger, "Leaving free_catched_result.");
}

void
CarobStmt::free_long_data(MYSQL_STMT *stmt)
{
  LOG4CXX_DEBUG(logger, "Entering free_long_data: stmt=" << stmt);

  if (stmt)
  {
    MYSQL_BIND *params=stmt->params;
    for (unsigned colNo=0; colNo < stmt->param_count && params; colNo++, params++)
      if (params->long_data_used)
      {
        params->long_data_used = 0;
#ifdef MYSQL5
        delete (string *)(params->row_ptr);
        params->row_ptr = 0;
#else
        delete (string *)(params->inter_buffer);
        params->inter_buffer = 0;
#endif
      }
  }

  LOG4CXX_DEBUG(logger, "Leaving free_long_data.");
}

void
CarobStmt::clear()
{
  LOG4CXX_DEBUG(logger, "Entering clear.");

  if (cmysql && c_stmt)
  {
    cmysql->connectionPtr->deleteStatement(c_stmt);
    c_stmt = 0;
  }

  liveResultSet = 0;

  free_catched_result(m_stmt);
  free_long_data(m_stmt);
  delete_query_fields(m_stmt->fields, m_stmt->field_count);
  
  FREE_AND_NULL_ARRAY(m_stmt->params);
  FREE_AND_NULL_ARRAY(m_stmt->bind);
  
  m_stmt->bind_param_done = false;
  m_stmt->bind_result_done = false;
  m_stmt->param_count = 0;
  m_stmt->field_count = 0;
  m_stmt->last_errno = 0;
  *m_stmt->last_error = 0;
  *m_stmt->sqlstate = 0;
  m_stmt->affected_rows = 0;

  LOG4CXX_DEBUG(logger, "Leaving clear.");
}

std::string
CarobStmt::from_wstring(const wstring &w)
{
  return cmysql ? cmysql->from_wstring(w) : string();
}

std::wstring
CarobStmt::to_wstring(const string &s)
{
  return cmysql ? cmysql->to_wstring(s) : wstring();
}
