/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon, Csaba Simon
 * Contributor(s): 
 */
 
#include "IniParser.hpp"

#include <string>
#include <fstream>

using namespace std;
 
bool IniParser::parseFile(const char *name)
{
  string group, line;
  bool in_group = false;

  ifstream file(name);
  
  if (file)
  {
	  while (getline(file, line))
	  {
	    removeSpacesAndComments(line);
	    if (line.empty())
	     continue;
	    if (parseGroup(line, group))
	    {
	      in_group = true;
	    }
	    else
	    {
	      if (in_group)
	        parseName(line, group);
	    }
	  }
    return true;
  }
  
  return false;
}
 
string IniParser::get(std::string group, std::string name) const
{
  std::map<std::string, std::map<std::string, std::string> >::const_iterator group_iter = keys.find(group);
  if (group_iter == keys.end())
    return "";

  std::map<std::string, std::string>::const_iterator name_iter = group_iter->second.find(name);
  if (name_iter == group_iter->second.end())
    return "";
  
	return name_iter->second;
}

void IniParser::removeLeadingSpaces(string& s)
{
  unsigned int i = 0;
  
  while (i != s.size() && isspace(s[i]))
    ++i;
    
  s.assign(s, i, s.size() - i);
}

void IniParser::removeTrailingSpaces(string& s)
{
  unsigned int i = s.size() > 0 ? s.size() - 1 : 0;
  
  while (i >= 0 && isspace(s[i]))
    --i;
    
  s.assign(s, 0, i + 1);
}

void IniParser::removeComments(string& line)
{
  unsigned int i = 0;
  bool in_quote = false;

	if (line[0] == '#' || line[0] == ';')
	{
		line.assign(line, 0, 0);
		return;
	}
	
  for (i = 0; i != line.size(); ++i)
    {
      if (line[i] == '\"')
      {
        in_quote = !in_quote;
      }
      else if (line[i] == '#' && !in_quote)
      {
        break;
      }
    }
    
    line.assign(line, 0, i);
}  

void IniParser::removeSpacesAndComments(string& line)
{
  removeLeadingSpaces(line);
  removeComments(line);  
  removeTrailingSpaces(line);
}

bool IniParser::parseGroup(string line, string& group)
{
  if (line[0] != '[' || line[line.size() -1] != ']')
    return false;
  
  group.assign(line, 1, line.size() - 2);
  removeLeadingSpaces(group);
  removeTrailingSpaces(group);
  
  return true;
}

void IniParser::parseName(string line, string group)
{
  unsigned int i = line.find('=');
  
  if (i == string::npos)
    return;
  
  string name(line, 0, i);
  removeTrailingSpaces(name);
  
  keys[group][name] = parseValue(string(line, i+1));
}

string IniParser::parseValue(string value)
{
  removeLeadingSpaces(value);
  
  return value;
}
